import numpy as np
from collections import Counter


def generate_matrix(infile, res, genome, dist, outdir, cell):
    fin = open(infile, 'r')
    if res >= 1000000:
        k = str(res / 1000000) + 'mb'
    elif res >= 1000:
        k = str(res / 1000) + 'kb'

    if genome[:2] == 'hg':
        chrom = [str(i + 1) for i in range(22)]
    elif genome[:2] == 'mm':
        chrom = [str(i + 1) for i in range(19)]

    count = Counter()
    for line in fin:
        tmp = line.strip().split('\t')
        tmp[1], tmp[5] = tmp[1][3:], tmp[5][3:]
        if (np.abs(int(tmp[6]) - int(tmp[2])) >= dist) and (tmp[1] == tmp[5]) and (tmp[1] in chrom):
            tmp[2], tmp[6] = int(tmp[2]) // res, int(tmp[6]) // res
            if tmp[2] > tmp[6]:
                tmp[2], tmp[6] = tmp[6], tmp[2]
            if tmp[2] != tmp[6]:
                pos = '-'.join([tmp[1], str(tmp[2]), str(tmp[6])])
                count[pos] += 1

    fin.close()
    fout = {c: open(outdir + 'chr' + c + '/' + cell + '_chr' + c + '.txt', 'w') for c in chrom}
    for key in count:
        tmp = key.split('-')
        fout[tmp[0]].write('{0}\t{1}\t{2}.0\n'.format(tmp[1], tmp[2], count[key]))

    for c in chrom:
        fout[c].close()
    return



# command time python /gale/ddn/snm3C/humanPFC/code/embed_mergechr.py --embed_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/100kb_resolution/filelist/embedlist_pad1_std1_rp0.5_sqrtvc.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/100kb_resolution/merged/pad1_std1_rp0.5_sqrtvc

import h5py
import argparse
import numpy as np
from sklearn.decomposition import TruncatedSVD

def embed_mergechr(embed_list, outprefix, dim=20, norm_sig=True):
	embedlist = np.loadtxt(embed_list, dtype=np.str)
	matrix_reduce = np.concatenate([np.load(x) for x in embedlist], axis=1)
	if norm_sig:
		matrix_reduce = matrix_reduce[1:] / matrix_reduce[0]
	else:
		matrix_reduce = matrix_reduce[1:]
	svd = TruncatedSVD(n_components=dim, algorithm='arpack')
	matrix_reduce = svd.fit_transform(matrix_reduce)
	if norm_sig:
		matrix_reduce = matrix_reduce / svd.singular_values_
	with h5py.File(f'{outprefix}.svd{dim}.hdf5', 'w') as f:
		tmp = f.create_dataset('data', matrix_reduce.shape, dtype='float32', compression='gzip')
		tmp[()] = matrix_reduce
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--embed_list', type=str, default=None, help='Full path of a file containing the full path to dimension reduction files of all chromosomes')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of final dimension reduction file including directory')
parser.add_argument('--dim', type=int, default=20, help='Number of dimensions to return from SVD')
parser.add_argument('--use_pc', dest='norm_sig', action='store_false', help='Not to normalize PCs by singular values')
parser.set_defaults(norm_sig=True)
opt = parser.parse_args()

embed_mergechr(opt.embed_list, opt.outprefix, opt.dim, opt.norm_sig)
'''



# command time python /gale/ddn/snm3C/humanPFC/code/comp_concatcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/filelist/cpglist_raw_chr${c}.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/merged/raw_chr${c} --ncpus 10

import argparse
import numpy as np
from multiprocessing import Pool

def load_cpg(i):
    data = np.load(celllist[i])
    return [i, data[:-3], data[-3:]]

def comp_concatcell_chr(cell_list, outprefix, ncpus=10):
	global celllist
	celllist = np.loadtxt(cell_list, dtype=np.str)
	p = Pool(ncpus)
	result = p.map(load_cpg, np.arange(len(celllist)))
	p.close()
	comp = np.zeros((len(celllist), len(result[0][1])))
	score = np.zeros((len(celllist), 3))
	for i,x,s in result:
		comp[i] = x.copy()
		score[i] = s.copy()
	np.save(f'{outprefix}.cpgcomp.npy', comp)
	np.save(f'{outprefix}.cpgcompstr.npy', score)

'''
parser = argparse.ArgumentParser()
parser.add_argument('--cell_list', type=str, default=None, help='Full path of a file containing the full path of all CpG npy files to be concatenate')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of concatenated matrix including directory')
parser.add_argument('--ncpus', type=int, default=10, help='# threads for parallelization')
opt = parser.parse_args()

comp_concatcell_chr(opt.cell_list, opt.outprefix, opt.ncpus)
'''


import pathlib
from concurrent.futures import ProcessPoolExecutor, as_completed
import numpy as np
import pandas as pd
from rpy2.rinterface_lib.embedded import RRuntimeError
from rpy2.robjects import r, pandas2ri, numpy2ri
from rpy2.robjects.packages import importr, isinstalled
from rpy2.robjects.vectors import StrVector
import cooler
from schicluster.cool.utilities import get_chrom_offsets
from scipy.sparse import csr_matrix, save_npz, load_npz, vstack
import anndata
import xarray as xr
import subprocess
import schicluster

pandas2ri.activate()
numpy2ri.activate()

PACKAGE_DIR = pathlib.Path(schicluster.__path__[0])


# check if some package is installed
def install_r_package(name):
    if not isinstalled(name):
        utils = importr('utils')
        utils.chooseCRANmirror(ind=1)
        utils.install_packages(StrVector([name]))


def domain_df_to_boundary(cool, total_results, resolution):
    # aggr to 1D boundary record,
    # 1 means either start or end of a domain,
    # 2 means both start and end of two domains
    bins = cool.bins()[:]
    chrom_offset = get_chrom_offsets(bins)
    use_results = total_results[total_results['name'] == 'domain']
    bin_loc = use_results.iloc[:, [1, 2]] // resolution
    bin_loc += use_results['chrom'].map(chrom_offset).values[:, None]
    bin_loc_data = np.zeros(bins.shape[0], dtype=np.int16)
    bin_loc_data[bin_loc['chromStart']] += 1
    bin_loc_data[bin_loc['chromEnd']] += 1
    bin_loc_data = csr_matrix(bin_loc_data)
    return bin_loc_data


def single_chrom_calculate_insulation_score(matrix, window_size=10):
    w = window_size
    score = np.ones(matrix.shape[0])
    for i in range(1, matrix.shape[0]):
        if i < w:
            intra = (matrix[:i, :i].sum() +
                     matrix[i:(i + w), i:(i + w)].sum()) / (i * (i + 1) / 2 +
                                                            (w * (w + 1) / 2))
            inter = matrix[:i, i:(i + w)].sum() / (i * (i + w))
        else:
            intra = (matrix[(i - w):i, (i - w):i].sum() +
                     matrix[i:(i + w), i:(i + w)].sum()) / (w * (w + 1))
            inter = matrix[(i - w):i, i:(i + w)].sum() / (w * w)
        score[i] = inter / (inter + intra)
    return score


def call_domain_and_insulation(cell_url,
                               output_prefix,
                               resolution=25000,
                               window_size=10):
    r.source(str(PACKAGE_DIR / 'domain/TopDom.R'))
    pandas2ri.activate()
    numpy2ri.activate()

    def run_top_dom(matrix, bins):
        j = matrix.indices + 1
        p = matrix.indptr
        x = matrix.data
        result = r.RunTopDom(j, p, x, bins, window_size)
        result = pd.DataFrame(result)
        return result

    cool = cooler.Cooler(cell_url)
    total_domain_results = []
    total_insulation_score = []
    for chrom in cool.chromnames:
        matrix = cool.matrix(balance=False, sparse=True).fetch(chrom).tocsc()
        bins = cool.bins().fetch(chrom).reset_index(drop=True)
        bins.columns = ["chr", "from.coord", "to.coord"]
        if (matrix.nnz < (matrix.shape[0] * matrix.shape[1] * 0.001)) or (matrix.shape[0] < 10):
            # skip if matrix is too less info or too small
            pass
        else:
            try:
                total_domain_results.append(run_top_dom(matrix, bins))
            except RRuntimeError:
                print('Got R error at', cell_url, chrom, matrix.shape, matrix.data.size, bins.shape)
        total_insulation_score.append(
            single_chrom_calculate_insulation_score(matrix, window_size))
    if len(total_domain_results) == 0:
        total_domain_results = pd.DataFrame([], columns=['chrom', 'chromStart', 'chromEnd', 'name'])
    else:
        total_domain_results = pd.concat(total_domain_results).reset_index(
            drop=True)
    domain_boundary = domain_df_to_boundary(cool, total_domain_results,
                                            resolution)
    insulation_score = np.concatenate(total_insulation_score)

    # save
    save_npz(f'{output_prefix}.boundary.npz', domain_boundary)
    np.savez(f'{output_prefix}.insulation.npz', insulation_score)
    return


def aggregate_boundary(cell_table, temp_dir, bins, output_path):
    total_boundary = []

    for cell_id, cell_url in cell_table.items():
        boundary_path = f'{temp_dir}/{cell_id}.boundary.npz'
        total_boundary.append(load_npz(boundary_path))
    total_boundary = vstack(total_boundary)
    adata = anndata.AnnData(X=total_boundary,
                            obs=pd.DataFrame([], index=cell_table.index.tolist()),
                            var=bins)
    adata.write_h5ad(output_path)
    return


def aggregate_insulation(cell_table, temp_dir, bins, output_path):
    total_insulation = []
    for cell_id, cell_url in cell_table.items():
        insulation_path = f'{temp_dir}/{cell_id}.insulation.npz'
        total_insulation.append(np.load(insulation_path)['arr_0'])
    total_insulation = np.vstack(total_insulation)
    total_insulation = pd.DataFrame(total_insulation,
                                    index=cell_table.index,
                                    columns=bins.index)
    total_insulation.index.name = 'cell'
    total_insulation.columns.name = 'bin'
    total_insulation = xr.DataArray(total_insulation)
    total_insulation.coords['bin_chrom'] = bins['chrom']
    total_insulation.coords['bin_start'] = bins['start']
    total_insulation.coords['bin_end'] = bins['end']
    total_insulation.to_netcdf(output_path)
    return


def multiple_call_domain_and_insulation(cell_table_path,
                                        output_prefix,
                                        resolution=25000,
                                        window_size=10,
                                        cpu=10):
    # install R package Matrix
    install_r_package('Matrix')

    cell_table = pd.read_csv(cell_table_path,
                             sep='\t',
                             index_col=0,
                             header=None,
                             squeeze=True)
    print(cell_table.shape[0], 'cells to calculate.')

    temp_dir = pathlib.Path(f'{output_prefix}_domain_temp')
    temp_dir.mkdir(exist_ok=True)

    # calculate individual cells
    with ProcessPoolExecutor(cpu) as exe:
        future_dict = {}
        for cell_id, cell_url in cell_table.items():
            cell_prefix = f'{temp_dir}/{cell_id}'
            if pathlib.Path(f'{cell_prefix}.insulation.npz').exists():
                continue
            future = exe.submit(call_domain_and_insulation,
                                cell_url,
                                cell_prefix,
                                resolution=resolution,
                                window_size=window_size)
            future_dict[future] = cell_id

        for future in as_completed(future_dict):
            cell_id = future_dict[future]
            print(f'{cell_id} finished.')
            future.result()

    # read bins from one of the cooler file
    # all cooler files should share the same bins
    cell_cool = cooler.Cooler(cell_table.iloc[0])
    bins = cell_cool.bins()[:]

    # aggregate boundary
    aggregate_boundary(cell_table=cell_table,
                       temp_dir=temp_dir,
                       bins=bins,
                       output_path=f'{output_prefix}.boundary.h5ad')

    # aggregate insulation
    aggregate_insulation(cell_table=cell_table,
                         temp_dir=temp_dir,
                         bins=bins,
                         output_path=f'{output_prefix}.insulation.nc')

    # cleanup
    subprocess.run(f'rm -rf {temp_dir}', shell=True)
    return



import logging
import pathlib

import cooler
import numpy as np
import pandas as pd


def get_chrom_offsets(bins_df):
    chrom_offset = {chrom: bins_df[bins_df['chrom'] == chrom].index[0]
                    for chrom in bins_df['chrom'].cat.categories}
    return chrom_offset


def write_coo(path, matrix, chunk_size=5000000):
    """
    Write chromosome contacts as chunked pixel df (cooler input)
    """
    matrix = matrix.tocoo(copy=False)
    df = pd.DataFrame({'bin1_id': matrix.row, 'bin2_id': matrix.col, 'count': matrix.data})
    with pd.HDFStore(path, complib='zlib', complevel=3) as hdf:
        if chunk_size is None:
            # no chunk
            hdf['c0'] = df
        else:
            for i, chunk_start in enumerate(range(0, df.shape[0], chunk_size)):
                hdf[f'c{i}'] = df.iloc[chunk_start:chunk_start + chunk_size]
    return


def chrom_iterator(input_dir, chrom_order, chrom_offset, chrom_wildcard='{chrom}.hdf'):
    for chrom in chrom_order:
        chrom_file = chrom_wildcard.format(chrom=chrom)
        output_path = f'{input_dir}/{chrom_file}'
        if not pathlib.Path(output_path).exists():
            continue
        with pd.HDFStore(output_path, 'r') as hdf:
            logging.debug(chrom)
            keys = {int(i[2:]): i for i in hdf.keys()}
            for i in sorted(keys.keys()):
                key = keys[i]
                chunk = hdf[key]
                chunk.iloc[:, :2] += chrom_offset[chrom]
                yield chunk


def aggregate_chromosomes(chrom_size_path,
                          resolution,
                          input_dir,
                          output_path,
                          chrom_wildcard='{chrom}.hdf'):
    chrom_sizes = cooler.read_chromsizes(chrom_size_path, all_names = True)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    cooler.create_cooler(cool_uri=output_path,
                         bins=bins_df,
                         pixels=chrom_iterator(input_dir=input_dir,
                                               chrom_order=bins_df['chrom'].unique(),
                                               chrom_offset=chrom_offset,
                                               chrom_wildcard=chrom_wildcard),
                         ordered=True,
                         dtypes={'count': np.float32})
    return


def cell_chunk(cell_url, chrom_sizes, chunk=50000000):
    cell_cool = cooler.Cooler(cell_url)
    chunk_df = cooler.binnify(chrom_sizes, chunk)
    for _, row in chunk_df.iterrows():
        chrom, start, end = row
        region = f'{chrom}:{start}-{end}'
        # this fetch selected a rectangle region, row is region, col is whole chrom
        data = cell_cool.matrix(balance=False, as_pixels=True).fetch(region, chrom)
        yield data


def aggregate_cells(output_path, cell_dir, chrom_size_path, resolution):
    chrom_sizes = cooler.read_chromsizes(chrom_size_path, all_names=True)
    bins_df = cooler.binnify(chrom_sizes, resolution)

    cell_pixel_dict = {
        cell_url.name.split('.')[0]: cell_chunk(str(cell_url), chrom_sizes=chrom_sizes)
        for cell_url in pathlib.Path(cell_dir).glob('*cool')
    }
    cooler.create_scool(output_path,
                        bins=bins_df,
                        cell_name_pixels_dict=cell_pixel_dict,
                        ordered=True,
                        mode='a')
    return



import pathlib
import schicluster
import cooler

PACKAGE_DIR = pathlib.Path(schicluster.__path__[0])


def prepare_impute(input_scool,
                   output_dir,
                   chrom_size_path,
                   output_dist,
                   window_size,
                   step_size,
                   resolution,
                   batch_size=100,
                   logscale=False,
                   pad=1,
                   std=1,
                   rp=0.5,
                   tol=0.01,
                   min_cutoff=1e-5,
                   cpu_per_job=10):
    """
    prepare snakemake files and directory structure for cell contacts imputation
    """
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)
    input_scool = str(pathlib.Path(input_scool).absolute())

    with open(PACKAGE_DIR / 'impute/impute.Snakefile') as f:
        snake_template = f.read()
    cell_list = cooler.fileops.list_coolers(input_scool)
    scool_cell_ids = [i.split('/')[-1] for i in cell_list]
    if logscale:
        logscale_str = '--logscale'
    else:
        logscale_str = ''

    chunk_dirs = []
    for i, chunk_start in enumerate(range(0, len(scool_cell_ids), batch_size)):
        chunk_dir = output_dir / f'chunk{i}'
        chunk_dir.mkdir(exist_ok=True)

        this_cell_ids = scool_cell_ids[chunk_start:chunk_start + batch_size]
        parameters = dict(
            input_scool=f"'{pathlib.Path(input_scool).absolute()}'",
            chrom_size_path=f"'{pathlib.Path(chrom_size_path).absolute()}'",
            logscale_str=f'"{logscale_str}"',
            pad=pad,
            std=std,
            window_size=int(window_size),
            step_size=int(step_size),
            resolution=int(resolution),
            output_dist=int(output_dist),
            rp=rp,
            tol=tol,
            min_cutoff=min_cutoff,
            cell_ids=str(this_cell_ids)
        )
        parameters_str = '\n'.join([f'{k} = {v}' for k, v in parameters.items()])
        this_snakefile = parameters_str + snake_template
        with open(output_dir / f'chunk{i}/Snakefile', 'w') as f:
            f.write(this_snakefile)
        chunk_dirs.append(chunk_dir)

    with open(output_dir / 'snakemake_cmd.txt', 'w') as f:
        for chunk_dir in chunk_dirs:
            cmd = f'snakemake -d {chunk_dir} --snakefile {chunk_dir}/Snakefile -j {cpu_per_job}'
            f.write(cmd + '\n')
    return



import subprocess
import numpy as np
import pandas as pd
from scipy.sparse import diags
import cooler
import pathlib
from concurrent.futures import ProcessPoolExecutor, as_completed
import xarray as xr


def get_cpg_profile(cell_url, fasta_path, hdf_output_path):
    temp_bed_path = f'{hdf_output_path}_bins.bed'
    temp_cpg_path = f'{hdf_output_path}_cpg_profile.txt'

    # count CpG per bin
    cool = cooler.Cooler(cell_url)
    bins = cool.bins()[:]
    bins.to_csv(temp_bed_path, index=None, header=None, sep='\t')
    subprocess.run(
        f'bedtools nuc -fi {fasta_path} -bed {temp_bed_path} -pattern CG -C > {temp_cpg_path}',
        shell=True)

    # get CpG ratio for each bin
    result = pd.read_csv(temp_cpg_path, sep='\t')
    result['ratio'] = (result['13_user_patt_count'] /
                       (result['12_seq_len'] - result['10_num_N'])).fillna(0)
    # chrom, start, end, CpG ratio
    result = result.iloc[:, [0, 1, 2, -1]]
    result.columns = ['chrom', 'start', 'end', 'cpg_ratio']
    result.to_hdf(hdf_output_path, key='data')

    # cleanup
    subprocess.run(f'rm -f {temp_bed_path} {temp_cpg_path}', shell=True)
    return


def compartment_strength(matrix, comp, cpg_ratio):
    bin_filter = cpg_ratio > 0

    # for calculating compartment strength
    tmp = comp[bin_filter]
    a_pos = (tmp > np.percentile(tmp, 80))
    b_pos = (tmp < np.percentile(tmp, 20))
    E = matrix.tocoo()
    decay = np.array([E.diagonal(i).mean() for i in range(E.shape[0])])
    E.data = E.data / decay[np.abs(E.col - E.row)]
    E = E.tocsr()[np.ix_(bin_filter, bin_filter)]
    aa = E[np.ix_(a_pos, a_pos)].sum()
    bb = E[np.ix_(b_pos, b_pos)].sum()
    ab = E[np.ix_(a_pos, b_pos)].sum()
    scores = np.array([aa, bb, ab])
    return scores


def single_chrom_compartment(matrix, cpg_ratio, calc_strength=False):
    matrix = matrix - diags(matrix.diagonal())  # remove diag
    matrix = matrix + matrix.T  # make full matrix
    matrix = matrix + diags((matrix.sum(axis=0).A.ravel() == 0).astype(int))

    # weighted average of cpg density
    comp = matrix.dot(cpg_ratio.values[:, None])[:, 0]

    if calc_strength:
        scores = compartment_strength(matrix, comp, cpg_ratio)
    else:
        scores = None
    return comp, scores


def single_cell_compartment(cell_url, cpg_profile_path, calc_strength,
                            output_prefix):
    cpg_profile = pd.read_hdf(cpg_profile_path)
    cool = cooler.Cooler(cell_url)

    all_comp = []
    scores = np.array([0, 0, 0])
    for chrom in cool.chromnames:
        cpg_ratio = cpg_profile.loc[cpg_profile['chrom'] == chrom, 'cpg_ratio']
        matrix = cool.matrix(balance=False, sparse=True).fetch(chrom)
        comp, scores = single_chrom_compartment(matrix,
                                                cpg_ratio,
                                                calc_strength=calc_strength)
        if calc_strength:
            scores += scores
        all_comp.append(comp)
    all_comp = np.concatenate(all_comp)
    if calc_strength:
        np.savez(f'{output_prefix}.comp.npz', all_comp, scores)
    else:
        np.savez(f'{output_prefix}.comp.npz', all_comp)
    return


def aggregate_compartment(cell_table, temp_dir, bins, output_path, calc_strength):
    total_comp = []
    total_scores = []
    for cell_id, cell_url in cell_table.items():
        comp_path = f'{temp_dir}/{cell_id}.comp.npz'
        npz_data = np.load(comp_path)
        total_comp.append(npz_data['arr_0'])
        if calc_strength:
            total_scores.append(npz_data['arr_1'])

    total_comp = np.vstack(total_comp)
    total_comp = pd.DataFrame(total_comp,
                              index=cell_table.index,
                              columns=bins.index)
    total_comp.index.name = 'cell'
    total_comp.columns.name = 'bin'
    total_comp = xr.Dataset({'compartment': total_comp})
    total_comp.coords['bin_chrom'] = bins['chrom']
    total_comp.coords['bin_start'] = bins['start']
    total_comp.coords['bin_end'] = bins['end']

    if calc_strength:
        total_scores = np.vstack(total_scores)
        total_scores = pd.DataFrame(total_scores,
                                    index=cell_table.index,
                                    columns=['AA', 'BB', 'AB'])
        total_scores.index.name = 'cell'
        total_scores.columns.name = 'sum'
    total_comp['summary'] = xr.DataArray(total_scores)
    total_comp.to_netcdf(output_path)
    return


def multiple_cell_compartment(cell_table_path,
                              output_prefix,
                              cpg_profile_path,
                              cpu=10,
                              calc_strength=False):
    cell_table = pd.read_csv(cell_table_path,
                             sep='\t',
                             index_col=0,
                             header=None,
                             squeeze=True)
    print(cell_table.shape[0], 'cells to calculate.')

    temp_dir = pathlib.Path(f'{output_prefix}_compartment_temp')
    temp_dir.mkdir(exist_ok=True)

    # calculate individual cells
    with ProcessPoolExecutor(cpu) as exe:
        future_dict = {}
        for cell_id, cell_url in cell_table.items():
            cell_prefix = f'{temp_dir}/{cell_id}'
            if pathlib.Path(f'{cell_prefix}.comp.npz').exists():
                continue
            future = exe.submit(single_cell_compartment,
                                cell_url=cell_url,
                                cpg_profile_path=cpg_profile_path,
                                output_prefix=cell_prefix,
                                calc_strength=calc_strength)
            future_dict[future] = cell_id

        for future in as_completed(future_dict):
            cell_id = future_dict[future]
            print(f'{cell_id} finished.')
            future.result()

    # read bins from one of the cooler file
    # all cooler files should share the same bins
    cell_cool = cooler.Cooler(cell_table.iloc[0])
    bins = cell_cool.bins()[:]

    # aggregate boundary
    aggregate_compartment(cell_table=cell_table,
                          temp_dir=temp_dir,
                          bins=bins,
                          output_path=f'{output_prefix}.compartment.nc',
                          calc_strength=calc_strength)

    # cleanup
    subprocess.run(f'rm -rf {temp_dir}', shell=True)
    return



import time
import h5py
import cv2
import numpy as np
from scipy.sparse import csr_matrix, save_npz, diags, eye
from scipy.sparse.linalg import norm

cv2.useOptimized()
hg38dim = [248956422, 242193529, 198295559, 190214555, 181538259, 170805979, 159345973, 145138636, 138394717, 133797422,
           135086622, 133275309, 114364328, 107043718, 101991189, 90338345, 83257441, 80373285, 58617616, 64444167,
           46709983, 50818468]
hg19dim = [249250621, 243199373, 198022430, 191154276, 180915260, 171115067, 159138663, 146364022, 141213431, 135534747,
           135006516, 133851895, 115169878, 107349540, 102531392, 90354753, 81195210, 78077248, 59128983, 63025520,
           48129895, 51304566]
mm10dim = [195471971, 182113224, 160039680, 156508116, 151834684, 149736546, 145441459, 129401213, 124595110, 130694993,
           122082543, 120129022, 120421639, 124902244, 104043685, 98207768, 94987271, 90702639, 61431566]


def random_walk_cpu(P, rp, tol, dist, spr):
    global ngene
    if rp == 1:
        return P
    I = eye(ngene)
    Q = rp * I + (1 - rp) * P
    if dist < ngene:
        idx = np.triu_indices(ngene, 0)
        idxfilter = ((idx[1] - idx[0]) < dist)
        idx = (
            np.concatenate((idx[0][idxfilter], idx[1][idxfilter])),
            np.concatenate((idx[1][idxfilter], idx[0][idxfilter])))
        mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])), P.shape)
    start_time = time.time()
    for i in range(30):
        Q_new = rp * I + (1 - rp) * P.dot(Q)
        delta = norm(Q - Q_new)
        Q = Q_new.copy()
        sparsity = Q.nnz / ngene / ngene
        if (dist < ngene) and (sparsity > spr):
            Q = Q.multiply(mask)
        end_time = time.time()
        print('Iter', i + 1, 'takes', end_time - start_time, 'seconds, loss', delta, 'sparsity', sparsity)
        if delta < tol:
            break
    return Q


def impute_cell(indir,
                outdir,
                cell,
                chrom,
                res,
                genome,
                mode,
                logscale=False,
                pad=1,
                std=1,
                rp=0.5,
                tol=0.01,
                rwr_dist=500000000,
                rwr_sparsity=1,
                output_dist=500000000,
                output_format='hdf'):
    if chrom[:3] == 'chr':
        c = chrom[3:]
    else:
        c = chrom
    if not mode:
        mode = 'pad' + str(pad) + '_std' + str(std) + '_rp' + str(rp) + '_sqrtvc'

    if genome == 'hg38':
        chrom = [str(i + 1) for i in range(22)]
        chromsize = {x: y for x, y in zip(chrom, hg38dim)}
    elif genome == 'hg19':
        chrom = [str(i + 1) for i in range(22)]
        chromsize = {x: y for x, y in zip(chrom, hg19dim)}
    elif genome == 'mm10':
        chrom = [str(i + 1) for i in range(19)]
        chromsize = {x: y for x, y in zip(chrom, mm10dim)}
    else:
        # TODO instead of using fixed chrom info, use chrom size file
        raise NotImplementedError

    start_time = time.time()
    ngene = int(chromsize[c] // res) + 1
    D = np.loadtxt(indir + cell + '_chr' + c + '.txt')
    if len(D) == 0:
        D = np.array([[0, 0, 0]])
    elif len(D.shape) == 1:
        D = D.reshape(1, -1)

    A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene))
    if logscale:
        A.data = np.log2(A.data + 1)

    end_time = time.time()
    print('Loading chr', c, 'takes', end_time - start_time, 'seconds')

    start_time = time.time()
    B = cv2.GaussianBlur((A + A.T).toarray().astype(np.float32), (pad * 2 + 1, pad * 2 + 1), std)
    end_time = time.time()
    print('Convolution chr', c, 'take', end_time - start_time, 'seconds')

    start_time = time.time()
    B = csr_matrix(B)
    B = B - diags(B.diagonal())
    B = B + diags((B.sum(axis=0).A.ravel() == 0).astype(int))
    d = diags(1 / B.sum(axis=0).A.ravel())
    P = d.dot(B)
    Q = random_walk_cpu(P, rp, tol, int(rwr_dist // res) + 1, rwr_sparsity)
    print('RWR', time.time() - start_time)

    start_time = time.time()
    E = Q + Q.T
    d = E.sum(axis=0).A.ravel()
    d[d == 0] = 1
    b = diags(1 / np.sqrt(d))
    E = b.dot(E).dot(b)
    print('SQRTVC', time.time() - start_time)

    start_time = time.time()
    idx = np.triu_indices(E.shape[0], 0)
    idxfilter = ((idx[1] - idx[0]) < (output_dist // res + 1))
    idx = (idx[0][idxfilter], idx[1][idxfilter])
    mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])), E.shape)
    E = E.multiply(mask)
    E = E.tocsr()
    print('Filter', time.time() - start_time)

    if output_format == 'npz':
        save_npz(outdir + cell + '_chr' + c + '_' + mode + '.npz', E)
    else:
        f = h5py.File(outdir + cell + '_chr' + c + '_' + mode + '.hdf5', 'w')
        g = f.create_group('Matrix')
        g.create_dataset('data', data=E.data, dtype='float32', compression='gzip')
        g.create_dataset('indices', data=E.indices, dtype=int, compression='gzip')
        g.create_dataset('indptr', data=E.indptr, dtype=int, compression='gzip')
        g.attrs['shape'] = E.shape
        f.close()
    return



# command time python /gale/ddn/snm3C/humanPFC/code/loop_mergechr.py --inprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_${mode}_dist_trim/L23_${mode}_dist_trim --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_${mode}_dist_trim/L23_${mode}_dist_trim --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes
# command time python /gale/ddn/snm3C/humanPFC/code/loop_mergechr.py --inprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_${mode}_dist_trim/L23_${mode}_dist_trim --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_${mode}_dist_trim/L23_${mode}_dist_trim.split --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes --split_file /gale/netapp/home/zhoujt/genome/hg19/hg19.chrsplit.bed

import time
import argparse
import numpy as np
import pandas as pd
from heapq import *
from statsmodels.sandbox.stats.multicomp import multipletests as FDR

def loop_mergechr(inprefix, outprefix, chrom_file, split_file=None, res=10000, 
				thres_bl=1.33, thres_d=1.33, thres_h=1.2, thres_v=1.2, fdr_thres=0.1, dist_thres=20000, size_thres=1):

	def find_summit(loop, dist_thres):

		start_time = time.time()
		cord = loop[['x1', 'y1']].values // res
		idx = np.argsort(cord[:, 0])
		neighbor = {i:[] for i in range(len(idx))}
		for i in range(len(idx)-1):
			tmp = cord[idx[i]]
			for j in range(i+1,len(idx)):
				if cord[idx[j], 0] - tmp[0] > dist_thres:
					break
				if np.abs(tmp[1] - cord[idx[j], 1]) <= dist_thres:
					neighbor[idx[i]].append(idx[j])
					neighbor[idx[j]].append(idx[i])
		# dist = np.max([pairwise_distances(loop['x1'].values.reshape(-1,1)//res), 
		#				pairwise_distances(loop['y1'].values.reshape(-1,1)//res)], axis=0)
		# print('Compute distances takes', time.time() - start_time, 'seconds')
		# start_time = time.time()
		# G = nx.from_numpy_array((dist <= dist_thres))
		print('Build graph takes', time.time() - start_time, 'seconds')

		start_time = time.time()

		nodescore = loop['E'].values
		flag = np.zeros(len(nodescore))
		tot = len(nodescore)
		summit = []
		nodeheap = (loop['E'] * -1).reset_index().reset_index()[['E', 'level_0']].values.tolist()
		heapify(nodeheap)

		while tot>0:
			t = int(heappop(nodeheap)[1])
			while flag[t]:
				t = int(heappop(nodeheap)[1])
			q = [t]
			flag[t] = 1
			tot -= 1
			head = 0
			flagtmp = np.zeros(len(nodescore))
			while (head < len(q)):
				for t in neighbor[q[head]]:
					if not flagtmp[t] and nodescore[t]<nodescore[q[head]]:
						if not flag[t]:
							flag[t] = 1
							tot -= 1
						flagtmp[t] = 1
						q.append(t)
				head += 1
			summit.append([q[0], len(q)])
		summit = np.array(summit)
		loop = loop.iloc[summit[:,0]]
		loop['size'] = summit[:,1]
		print('BFS takes', time.time() - start_time, 'seconds')
		return loop

	chrom = np.loadtxt(chrom_file, dtype=np.str)[:,0]
	if not split_file:
		chrom_split = chrom.copy() 
	else:
		splitbed = pd.read_csv(split_file, sep='\t', header=None, index_col=0)
		chrom_split = np.concatenate([[c+'p', c+'q'] if c in splitbed.index else [c] for c in chrom])
	chrom_split = np.array([c if c[:3]=='chr' else 'chr'+c for c in chrom_split])

	start_time = time.time()
	loopall = []
	for c in chrom_split:
		data = pd.read_hdf(f'{inprefix}_{c}.loop.hdf5', key='loop')
		data['bkfilter'] = (((data['E']/data['E_bl'] > thres_bl) | (data['E_bl']<0)) & ((data['E']/data['E_donut'] > thres_d) | (data['E_donut']<0)) & ((data['E']/data['E_h'] > thres_h) | (data['E_h']<0)) & ((data['E']/data['E_v'] > thres_v) | (data['E_v']<0)))
		data['x1'] = data['x1'].astype(int) * res
		data['y1'] = data['y1'].astype(int) * res
		data['x2'] = data['x1'] + res
		data['y2'] = data['y1'] + res
		if c[-1]=='p' or c[-1]=='q':
			data['chr'] = c[:-1]
			if c[-1]=='q':
				data[['x1','x2','y1','y2']] += splitbed.loc[c[:-1],2] // res * res
		else:
			data['chr'] = c
		loopall.append(data)

	loopall = pd.concat(loopall, axis=0)
	loopall['rFDR'] = FDR(loopall['rpv'], 0.1, 'fdr_bh')[1]
	loopall['tFDR'] = FDR(loopall['tpv'], 0.1, 'fdr_bh')[1]
	print('Merge chromosomes takes', time.time() - start_time, 'seconds')

	loop = loopall.loc[(loopall['bkfilter']==1) & (loopall['tFDR']<fdr_thres)]
	loop.sort_values(by=['chr', 'x1', 'y1'])[['chr', 'x1', 'x2', 'chr', 'y1', 'y2', 'E']].to_csv(f'{outprefix}.loop.bedpe', sep='\t', index=False, header=None)
	scloop = loopall.loc[loopall['tFDR']<fdr_thres]
	scloop.sort_values(by=['chr', 'x1', 'y1'])[['chr', 'x1', 'x2', 'chr', 'y1', 'y2', 'E']].to_csv(f'{outprefix}.scloop.bedpe', sep='\t', index=False, header=None)
	bkloop = loopall.loc[loopall['bkfilter']==1]
	bkloop.sort_values(by=['chr', 'x1', 'y1'])[['chr', 'x1', 'x2', 'chr', 'y1', 'y2', 'E']].to_csv(f'{outprefix}.bkloop.bedpe', sep='\t', index=False, header=None)

	summit = pd.concat([find_summit(loop[loop['chr']==c], dist_thres//res) for c in chrom], axis=0)
	suumit = summit[summit['size'] >= size_thres]
	summit.sort_values(by=['chr', 'x1', 'y1'])[['chr', 'x1', 'x2', 'chr', 'y1', 'y2', 'E', 'size']].to_csv(f'{outprefix}.loopsummit.bedpe', sep='\t', index=False, header=None)

	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--inprefix', type=str, default=None, help='Full path of a file containing the full path of all imputed files to be merged without .hdf5 suffix')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of merged matrix including directory')
parser.add_argument('--chrom_file', type=str, default=None, help='Path to the chromosome size files containing all chromosome to be analyzed as the first column') ## mm10 or hg38
parser.add_argument('--split_file', type=str, default=None, help='Path to the bed file containing all chromosomes need to split and one region per chromosome as splitting point')
parser.add_argument('--res', type=int, default=10000, help='Bin size as integer to generate contact matrix')
parser.add_argument('--thres_bl', type=int, default=1.33, help='Lowest fold change threshold against bottom left background')
parser.add_argument('--thres_d', type=int, default=1.33, help='Lowest fold change threshold against donut background')
parser.add_argument('--thres_h', type=int, default=1.2, help='Lowest fold change threshold against horizontal background')
parser.add_argument('--thres_v', type=int, default=1.2, help='Lowest fold change threshold against vertical background')
parser.add_argument('--fdr_thres', type=int, default=0.1, help='Highest t-test FDR threshold of loops')
parser.add_argument('--dist_thres', type=int, default=20000, help='Highest distance threshold to merge loops into summit')
parser.add_argument('--size_thres', type=int, default=1, help='Lowest loop number threshold of summit')
opt = parser.parse_args()

loop_mergechr(opt.inprefix, opt.outprefix, opt.chrom_file, opt.split_file, opt.res, 
		opt.thres_bl, opt.thres_d, opt.thres_h, opt.thres_v, opt.fdr_thres, opt.dist_thres, opt.size_thres)
'''
